/*****************************************************************************
 * Open MCT, Copyright (c) 2014-2018, United States Government
 * as represented by the Administrator of the National Aeronautics and Space
 * Administration. All rights reserved.
 *
 * Open MCT is licensed under the Apache License, Version 2.0 (the
 * "License"); you may not use this file except in compliance with the License.
 * You may obtain a copy of the License at
 * http://www.apache.org/licenses/LICENSE-2.0.
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
 * WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
 * License for the specific language governing permissions and limitations
 * under the License.
 *
 * Open MCT includes source code licensed under additional open source
 * licenses. See the Open Source Licenses file (LICENSES.md) included with
 * this source code distribution or the Licensing information page available
 * at runtime from the About dialog for additional information.
 *****************************************************************************/

define([
    'moment',
    'zepto',
    '../../res/templates/snapshotTemplate.html',
    'vue',
    'painterro'
],
function (
    Moment,
    $,
    SnapshotTemplate,
    Vue,
    Painterro
) {
    function EmbedController (openmct, domainObject) {
        this.openmct = openmct;
        this.domainObject = domainObject;
        this.objectService = openmct.$injector.get('objectService');
        this.navigationService = openmct.$injector.get('navigationService');
        this.popupService = openmct.$injector.get('popupService');
        this.agentService = openmct.$injector.get('agentService');
        this.dialogService = openmct.$injector.get('dialogService');


        this.navigate = this.navigate.bind(this);
        this.exposedData = this.exposedData.bind(this);
        this.exposedMethods = this.exposedMethods.bind(this);
        this.toggleActionMenu = this.toggleActionMenu.bind(this);
    }

    EmbedController.prototype.navigate = function (embedType) {
        this.objectService.getObjects([embedType]).then(function (objects) {
            this.navigationService.setNavigation(objects[embedType]);
        }.bind(this));
    };

    EmbedController.prototype.openSnapshot = function (domainObject, entry, embed) {

        function annotateSnapshot(openmct) {
            return function () {

                var save = false,
                    painterroInstance = {},
                    annotateOverlay = new Vue({
                        template: '<div id="snap-annotation"></div>'
                    }),
                    self = this;

                openmct.overlays.overlay({
                    element: annotateOverlay.$mount().$el,
                    size: 'large',
                    buttons: [
                        {
                            label: 'Cancel',
                            callback: function () {
                                save = false;
                                painterroInstance.save();
                            }
                        },
                        {
                            label: 'Save',
                            callback: function () {
                                save = true;
                                painterroInstance.save();
                            }
                        }
                    ],
                    onDestroy: function () {
                        annotateOverlay.$destroy(true);
                    }
                });

                painterroInstance = Painterro({
                    id: 'snap-annotation',
                    activeColor: '#ff0000',
                    activeColorAlpha: 1.0,
                    activeFillColor: '#fff',
                    activeFillColorAlpha: 0.0,
                    backgroundFillColor: '#000',
                    backgroundFillColorAlpha: 0.0,
                    defaultFontSize: 16,
                    defaultLineWidth: 2,
                    defaultTool: 'ellipse',
                    hiddenTools: ['save', 'open', 'close', 'eraser', 'pixelize', 'rotate', 'settings', 'resize'],
                    translation: {
                        name: 'en',
                        strings: {
                            lineColor: 'Line',
                            fillColor: 'Fill',
                            lineWidth: 'Size',
                            textColor: 'Color',
                            fontSize: 'Size',
                            fontStyle: 'Style'
                        }
                    },
                    saveHandler: function (image, done) {
                        if (save) {
                            var entryPos = self.findInArray(domainObject.entries, entry.id),
                                embedPos = self.findInArray(entry.embeds, embed.id);

                            if (entryPos !== -1 && embedPos !== -1) {
                                var url = image.asBlob(),
                                    reader = new window.FileReader();

                                reader.readAsDataURL(url);
                                reader.onloadend = function () {
                                    var snapshot = reader.result,
                                        snapshotObject = {
                                            src: snapshot,
                                            type: url.type,
                                            size: url.size,
                                            modified: Date.now()
                                        },
                                        dirString = 'entries[' + entryPos + '].embeds[' + embedPos + '].snapshot';

                                    openmct.objects.mutate(domainObject, dirString, snapshotObject);
                                };
                            }
                        } else {
                            console.log('You cancelled the annotation!!!');
                        }
                        done(true);
                    }
                }).show(embed.snapshot.src);
            };
        }

        var self = this,
            snapshot = new Vue({
                template: SnapshotTemplate,
                data: function () {
                    return {
                        embed: self.embed
                    };
                },
                methods: {
                    formatTime: self.formatTime,
                    annotateSnapshot: annotateSnapshot(self.openmct),
                    findInArray: self.findInArray
                }
            });

        function onDestroyCallback() {
            snapshot.$destroy(true);
        }

        var snapshotOverlay = this.openmct.overlays.overlay({
            element: snapshot.$mount().$el,
            onDestroy: onDestroyCallback,
            size: 'large',
            buttons: [
                {
                    label: 'Done',
                    emphasis: true,
                    callback: function () {
                        snapshotOverlay.dismiss();
                    }
                }
            ]
        });
    };

    EmbedController.prototype.formatTime = function (unixTime, timeFormat) {
        return Moment(unixTime).format(timeFormat);
    };

    EmbedController.prototype.findInArray = function (array, id) {
        var foundId = -1;

        array.forEach(function (element, index) {
            if (element.id === id) {
                foundId = index;
                return;
            }
        });

        return foundId;
    };

    EmbedController.prototype.actionToMenuDecorator = function (action) {
        return {
            name: action.getMetadata().name,
            cssClass: action.getMetadata().cssClass,
            perform: action.perform
        };
    };

    EmbedController.prototype.populateActionMenu = function (objectService, actionService) {
        return function () {
            var self = this;

            objectService.getObjects([self.embed.type]).then(function (resp) {
                var domainObject = resp[self.embed.type],
                    previewAction = actionService.getActions({key: 'mct-preview-action', domainObject: domainObject})[0];

                self.actions.push(self.actionToMenuDecorator(previewAction));
            });
        };
    };

    EmbedController.prototype.removeEmbedAction = function () {
        var self = this;

        return {
            name: 'Remove Embed',
            cssClass: 'icon-trash',
            perform: function (embed, entry) {
                var entryPosition = self.findInArray(self.domainObject.entries, entry.id),
                    embedPosition = self.findInArray(entry.embeds, embed.id);

                var dialog = self.openmct.overlays.dialog({
                    iconClass: "alert",
                    message: 'This Action will permanently delete this embed. Do you wish to continue?',
                    buttons: [{
                        label: "No",
                        callback: function () {
                            dialog.dismiss();
                        }
                    },
                    {
                        label: "Yes",
                        emphasis: true,
                        callback: function () {
                            entry.embeds.splice(embedPosition, 1);
                            var dirString = 'entries[' + entryPosition + '].embeds';

                            self.openmct.objects.mutate(self.domainObject, dirString, entry.embeds);
                            dialog.dismiss();
                        }
                    }]
                });
            }
        };
    };

    EmbedController.prototype.toggleActionMenu = function (event) {
        event.preventDefault();

        var body = $(document.body),
            container = $(event.target.parentElement.parentElement),
            initiatingEvent = this.agentService.isMobile() ?
                'touchstart' : 'mousedown',
            menu = container.find('.menu-element'),
            dismissExistingMenu;

        // Remove the context menu
        function dismiss() {
            container.find('.hide-menu').append(menu);
            body.off(initiatingEvent, dismiss);
            menu.off(initiatingEvent, menuClickHandler);
            dismissExistingMenu = undefined;
        }

        function menuClickHandler(e) {
            e.stopPropagation();
            window.setTimeout(dismiss, 300);
        }

        // Dismiss any menu which was already showing
        if (dismissExistingMenu) {
            dismissExistingMenu();
        }

        // ...and record the presence of this menu.
        dismissExistingMenu = dismiss;

        this.popupService.display(menu, [event.pageX,event.pageY], {
            marginX: 0,
            marginY: -50
        });

        // Stop propagation so that clicks or touches on the menu do not close the menu
        menu.on(initiatingEvent, menuClickHandler);

        body.on(initiatingEvent, dismiss);

    };

    EmbedController.prototype.exposedData = function () {
        return {
            actions: [this.removeEmbedAction()],
            showActionMenu: false
        };
    };

    EmbedController.prototype.exposedMethods = function () {
        var self = this;

        return {
            navigate: self.navigate,
            openSnapshot: self.openSnapshot,
            formatTime: self.formatTime,
            toggleActionMenu: self.toggleActionMenu,
            actionToMenuDecorator: self.actionToMenuDecorator,
            findInArray: self.findInArray
        };
    };

    return EmbedController;
});
